; File:     Timing.inc
; ----------------------------------------------------------------------------------------------
;                   The following macros was written by MichaelW
;                      and we may found it in timers.asm
; ----------------------------------------------------------------------------------------------
START_COUNTER_CYCLE_HIGH_PRIORITY_CLASS   MACRO         ;LoopCount, Priority
                      LOCAL   label

            IFNDEF __counter__stuff__defined__
                __counter__stuff__defined__ equ <1>
            .data
            ALIGN 8             ;; Optimal alignment for QWORD
            ; __counter__qword__count__  dq 0
            ; __counter__loop__count__   dd 0
            ; __counter__loop__counter__ dd 0

                _CounterQword   dq 0            ; __counter__qword__count__
                _LoopCount      dd 0            ; __counter__loop__count__
                _LoopCounter    dd 0            ; __counter__loop__counter__
                                
            .code
            ENDIF

            mov     _LoopCount, 1000000     ;LoopCount
            
            invoke  GetCurrentProcess
            invoke  SetPriorityClass, eax, HIGH_PRIORITY_CLASS  ;Priority

            xor     eax, eax        ;; Use same CPUID input value for each call
            cpuid                   ;; Flush pipe & wait for pending ops to finish
            rdtsc                   ;; Read Time Stamp Counter

            push    edx             ;; Preserve high-order 32 bits of start count
            push    eax             ;; Preserve low-order 32 bits of start count
            
            mov     _LoopCounter, 1000000   ;LoopCount
            xor     eax, eax
            cpuid                   ;; Make sure loop setup instructions finish
          ALIGN 16                  ;; Optimal loop alignment for P6
          @@:                       ;; Start an empty reference loop
            sub     _LoopCounter, 1
            jnz     short @B

            xor     eax, eax
            cpuid                   ;; Make sure loop instructions finish
            rdtsc                   ;; Read end count
            pop     ecx             ;; Recover low-order 32 bits of start count
            sub     eax, ecx        ;; Low-order 32 bits of overhead count in EAX
            pop     ecx             ;; Recover high-order 32 bits of start count
            sbb     edx, ecx        ;; High-order 32 bits of overhead count in EDX
            push    edx             ;; Preserve high-order 32 bits of overhead count
            push    eax             ;; Preserve low-order 32 bits of overhead count

            xor     eax, eax
            cpuid
            rdtsc
            push    edx             ;; Preserve high-order 32 bits of start count
            push    eax             ;; Preserve low-order 32 bits of start count
            mov     _LoopCounter, 1000000   ;LoopCount
            xor     eax, eax
            cpuid                   ;; Make sure loop setup instructions finish
ALIGN 16                            ;; Optimal loop alignment for P6
label:                              ;; Start test loop
            START_LOOP_TEST equ <label>
ENDM
; ----------------------------------------------------------------------------------------------
SAVE_COUNTER_CYCLE  MACRO  arg
                    LOCAL  $tmpstr$

        sub     _LoopCounter, 1
        jnz     START_LOOP_TEST

        xor   eax, eax
        cpuid                 ;; Make sure loop instructions finish
        rdtsc                 ;; Read end count
        pop   ecx             ;; Recover low-order 32 bits of start count
        sub   eax, ecx        ;; Low-order 32 bits of test count in EAX
        pop   ecx             ;; Recover high-order 32 bits of start count
        sbb   edx, ecx        ;; High-order 32 bits of test count in EDX
        pop   ecx             ;; Recover low-order 32 bits of overhead count
        sub   eax, ecx        ;; Low-order 32 bits of adjusted count in EAX
        pop   ecx             ;; Recover high-order 32 bits of overhead count
        sbb   edx, ecx        ;; High-order 32 bits of adjusted count in EDX

        mov   DWORD PTR _CounterQword, eax
        mov   DWORD PTR _CounterQword + 4, edx

        invoke GetCurrentProcess
        invoke SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

        finit
        fild  _CounterQword
        fild  _LoopCount
        fdiv
        fistp _CounterQword

        mov     eax, DWORD PTR _CounterQword
        ; ----------------------
        ;      Save time
        ; ----------------------
        call    SaveTiming
        
        print   str$(eax)
;---------------------------------------------------        
        $tmpstr$  CATSTR <chr$(">, <arg>, <",13,10)>        
        print   $tmpstr$, 13,10
ENDM
; 
START_COUNTER_TIMER_HIGH_PRIORITY_CLASS MACRO       ;loopcount:REQ, priority
            LOCAL label

        IFNDEF __timer__stuff__defined__
          __timer__stuff__defined__ equ <1>
          .data
          ALIGN 8                                   ;; Optimal alignment for QWORD
            __timer__pc__frequency__    dq 0
            __timer__pc__count__        dq 0
            __timer__loop__counter__    dd 0
            __timer__dw_count__         dd 0

          .code
        ENDIF

        invoke QueryPerformanceFrequency, ADDR __timer__pc__frequency__
        .IF eax != 0

            IFNB <priority>
              invoke GetCurrentProcess
              invoke SetPriorityClass, eax, HIGH_PRIORITY_CLASS     ;priority
            ENDIF

            invoke QueryPerformanceCounter, ADDR __timer__pc__count__
            push    DWORD PTR __timer__pc__count__ + 4
            push    DWORD PTR __timer__pc__count__

            mov   __timer__loop__counter__, 1000000                 ;loopcount
          ALIGN 16              ;; Optimal loop alignment for P6
          @@:                   ;; Start an empty reference loop
            sub   __timer__loop__counter__, 1
            jnz   @B

            invoke QueryPerformanceCounter, ADDR __timer__pc__count__

            pop   ecx           ;; Recover low-order 32 bits of start count
            sub   DWORD PTR __timer__pc__count__, ecx
            pop   ecx           ;; Recover high-order 32 bits of start count
            sbb   DWORD PTR __timer__pc__count__ + 4, ecx

            push    DWORD PTR __timer__pc__count__ + 4 ;; Overhead count
            push    DWORD PTR __timer__pc__count__     ;; Overhead count

            invoke QueryPerformanceCounter, ADDR __timer__pc__count__

            push    DWORD PTR __timer__pc__count__ + 4 ;; Start count
            push    DWORD PTR __timer__pc__count__     ;; Start count

            mov   __timer__loop__counter__, 1000000             ;loopcount
          ALIGN 16                             ;; Optimal loop alignment for P6
label:                                         ;; Start test loop
            __timer__loop__label__ equ <label>
        .ENDIF
    ENDM
; ----------------------------------------------------------------------------------------------
    SAVE_COUNTER_TIMER  MACRO  arg
                        LOCAL  $tmpstr$
    
        sub   __timer__loop__counter__, 1
        jnz   __timer__loop__label__

        invoke QueryPerformanceFrequency, ADDR __timer__pc__frequency__
        .IF eax != 0

            invoke QueryPerformanceCounter, ADDR __timer__pc__count__
            pop   ecx           ;; Recover low-order 32 bits of start count
            sub   DWORD PTR __timer__pc__count__, ecx
            pop   ecx           ;; Recover high-order 32 bits of start count
            sbb   DWORD PTR __timer__pc__count__ + 4, ecx
            pop   ecx           ;; Recover low-order 32 bits of overhead count
            sub   DWORD PTR __timer__pc__count__, ecx
            pop   ecx           ;; Recover high-order 32 bits of overhead count
            sbb   DWORD PTR __timer__pc__count__ + 4, ecx

            invoke GetCurrentProcess
            invoke SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

            finit
            fild  __timer__pc__count__
            fild  __timer__pc__frequency__
            fdiv
            mov   __timer__dw_count__, 1000
            fild  __timer__dw_count__
            fmul
            fistp __timer__dw_count__

            mov     eax, __timer__dw_count__
            ; ----------------------
            ;      Save time
            ; ----------------------
            call    SaveTiming
            
            print   str$(eax)            
;-----------------------------------------------        
            $tmpstr$  CATSTR <chr$(">, <arg>, <",13,10)>        
            print   $tmpstr$          

        .ELSE
          xor   eax, eax                            ;; No performance counter        
          print   "No performance counter",13,10
        .ENDIF
    ENDM
; 
.data
            dd 0   
_TblTiming0 dd 60 dup (?)

.code
SaveTiming  proc
            
            mov     edx, offset _TblTiming0
            mov     ecx, [edx-4]
            mov     [edx+ecx*4], eax
            add     ecx, 1
            mov     [edx-4], ecx            
            
            ret
SaveTiming  endp
;***********************************************************************************************
SortCycles  proc
            push    ebp
            push    esi
            push    edi

            mov     esi, offset _TblTiming0
            mov     edi, offset _TblString0            
            mov     ecx, [edi-4]
            
_nextECX:   sub     ecx, 1
            jns     short @F

            call    PrintTime

            pop     edi
            pop     esi            
            pop     ebp
            ret
            
    @@:     mov     edx, ecx
            
            mov     ebp, [esi+ecx*4]
            
_nextEDX:   sub     edx, 1
            js      _nextECX
            ;           
            mov     eax, [esi+edx*4]
            cmp     ebp, eax
            jbe     _nextEDX
            
            ; ---------------
            ; exchange timing
            ; ---------------
            mov     [esi+ecx*4], eax
            mov     [esi+edx*4], ebp
            mov     ebp, eax

            ; ---------------
            ; exchange string
            ; ---------------
            
            mov     ebx, [edi+ecx*4]
            mov     eax, [edi+edx*4]
            ;
            mov     [edi+ecx*4], eax
            mov     [edi+edx*4], ebx            
            
            jmp     _nextEDX
SortCycles  endp            
; ----------------------------------------------------------------------------------------------
PrintTime   proc
            
            print   " ", 13,10
            print   " ***** Time table *****",13,10,13,10
            mov     esi, offset _TblTiming0
            mov     edi, offset _TblString0
            mov     ebx, [edi-4]
            jmp     _loop
            
    @@:     print   str$([esi+ebx*4])
            print   [edi+ebx*4],13,10

    _loop:  sub     ebx, 1
            jns     short @B
            ret
PrintTime   endp
;***********************************************************************************************

